#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/syscall.h>
#include <linux/perf_event.h>
#include <errno.h>
#include <stdbool.h>
#include "cpucycles.h"
/*
 cpucycle() routines are based on the code in the supercop benchmark suite
 which is under public domain.

 The perf implementation is based on code by Austin Seipp
 and provided under the MIT License.

*/


cc_set *cc_set_init(int size) {
	cc_set *p = (cc_set *)malloc(sizeof(cc_set));
	p->array = (uint64_t *) malloc(sizeof(uint64_t)*size);
	p->size = size;

	return p;
}

void cc_set_free(cc_set *p) {
	free(p->array);
	free(p);
}

static int compare_uint64(const void *a, const void *b) {
	if (*(uint64_t *)a < *(uint64_t *)b) {
		return -1;
	} else if (*(uint64_t *)a > *(uint64_t *)b) {
		return 1;
	} else {
		return 0;
	}
}


void sort(cc_set *c) {
	qsort(c->array, c->size, sizeof(uint64_t), compare_uint64);
}

uint64_t median(cc_set *c) {
	sort(c);
	return c->array[c->size/2];
}

uint64_t min(cc_set *c) {
	sort(c);
	return c->array[0];
}

uint64_t max(cc_set *c) {
	sort(c);
	return c->array[c->size-1];
}

#if defined(USE_PERF)
int perfdev = -1;

void perf_init(void) {
	static struct perf_event_attr attr;
	attr.type = PERF_TYPE_HARDWARE;
	attr.config = PERF_COUNT_HW_CPU_CYCLES;
	perfdev = syscall(__NR_perf_event_open, &attr, 0, -1, -1, 0);

	if (perfdev > 0) {
		return;
	}

	printf("-perf_event_open() with PERF_COUNT_HW_CPU_CYCLES failed: %s\n", strerror(errno));
	exit(EXIT_FAILURE);
}

void perf_fin(void) {
	close(perfdev);
}

uint64_t cpucycles(void) {
	uint64_t result;
	if (read(perfdev, &result, sizeof(result)) == -1 ) {
		printf("read() with PERF_COUNT_HW_CPU_CYCLES failed: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	return result;
}

#elif defined(__i386__)
uint64_t cpucycles( void ) {
  uint64_t result;
  __asm__ __volatile__
  (
    ".byte 15;.byte 49"
    : "=A" ( result )
  );
  return result;
}
#elif defined(__x86_64__)
uint64_t cpucycles( void ) {
  uint64_t result;
  __asm__ __volatile__
  (
    ".byte 15;.byte 49\n"
    "shlq $32,%%rdx\n"
    "orq %%rdx,%%rax"
    : "=a" ( result ) ::  "%rdx"
  );
  return result;
}
#elif defined(__arm__)
#error broken, needs kernel module, try -DUSE_PERF
static void readticks(unsigned int *result) {
  static int enabled = 0;
  struct timeval t;
  unsigned int cc;
  if (!enabled) {
    asm volatile("mcr p15, 0, %0, c9, c12, 0" :: "r"(17));
    asm volatile("mcr p15, 0, %0, c9, c12, 1" :: "r"(0x8000000f));
    asm volatile("mcr p15, 0, %0, c9, c12, 3" :: "r"(0x8000000f));
    enabled = 1;
  }
  asm volatile("mrc p15, 0, %0, c9, c13, 0" : "=r"(cc));
  gettimeofday(&t,(struct timezone *) 0);
  result[0] = cc;
  result[1] = t.tv_usec;
  result[2] = t.tv_sec;
}

uint64_t cpucycles(void) {
  static int prev[3];
  static int now[3];
  static int64_t cyclespersec = 0;
  static uint64_t prevcycles = 0;
  uint64_t delta4;
  int deltan;
  int deltas;
  uint64_t guesscycles;
 
  readticks(now);
  delta4 = (unsigned int) (now[0] - prev[0]); /* unsigned change in number of cycles mod 2^32 */
  deltan = now[1] - prev[1]; /* signed change in number of nanoseconds mod 10^9 */
  deltas = now[2] - prev[2]; /* signed change in number of seconds */
  if ((deltas == 0 && deltan < 200000) || (deltas == 1 && deltan < -800000))
    return (prevcycles + delta4) * 1;

  prev[0] = now[0];
  prev[1] = now[1];
  prev[2] = now[2];

  if ((deltas == 0 && deltan < 300000) || (deltas == 1 && deltan < -700000)) {
    // actual number of cycles cannot have increased by 2^32 in <0.3ms
    cyclespersec = 1000000 * (uint64_t) delta4;
    cyclespersec /= deltan + 1000000 * (int64_t) deltas;
  } else {
    guesscycles = deltas * cyclespersec;
    guesscycles += (deltan * cyclespersec) / 1000000;
    while (delta4 + 2147483648ULL < guesscycles) delta4 += 4294967296ULL;
    /* XXX: could do longer-term extrapolation here */
  }

  prevcycles += delta4;
  return prevcycles;
}

#else
#error Your architecture is not yet supported
#endif

