#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/syscall.h>
#include <linux/perf_event.h>
#include <errno.h>
#include <gsl/gsl_histogram.h>
#include "cpucycles.h"

void run_100k_cycles() {
#if defined(__x86_64__)
	asm volatile (
		"pushq %rax;"
		"movq $33332, %rax;"
		"loop_100k: dec %rax;"
		"test %rax, %rax;"
		"jnz loop_100k;"
		"popq %rax;"
	);
#elif defined(__i386__)
	asm volatile (
		"push %eax;"
		"mov $33332, %eax;"
		"loop_100k: dec %eax;"
		"test %eax, %eax;"
		"jnz loop_100k;"
		"pop %eax;"
	);
#elif defined(__arm__) 
	asm volatile (
		"push {r1};"
		"ldr r1, =33331;"
		"loop_100k: subs %r1, #1;"
		"cmp r1, #0;"
		"bne loop_100k;"
		"pop {r1};"
	);

#endif


}

void test_cpucycles_function(uint64_t (get_cpucycles)(), char *ident) {
	uint64_t i, t, iterations = 100000;
	cc_set *set;

	char filename[300];
	snprintf((char *)&filename, 300, "%s-histogram.csv", ident);

	set = cc_set_init(iterations);
	for (i = 0; i < iterations; i++){
		t = get_cpucycles();
		run_100k_cycles();
		set->array[i] =  get_cpucycles() - t;
	}

	printf("%s - median: %llu \n",ident, median(set));

	double x;
    gsl_histogram * h = gsl_histogram_alloc(1000);
    gsl_histogram_set_ranges_uniform (h, min(set)-1, max(set)+1);

	for (i = 0; i < set->size; i++) {
		gsl_histogram_increment(h, (double)set->array[i]);
	}
	
	FILE *f = fopen(filename, "w");
	gsl_histogram_fprintf(f, h, "%g", "%g");
	gsl_histogram_free(h);
	fclose(f);
	cc_set_free(set);
}

int perfdev;
uint64_t cpucycles_perf() {
	uint64_t result = 0;
	if (read(perfdev, &result, sizeof(result)) < sizeof(result)) {
		printf("read() with PERF_COUNT_HW_CPU_CYCLES failed: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	return result;
}

void test_PERF_COUNT_HW_CPU_CYCLES() {
	static struct perf_event_attr attr;
	uint64_t before, after;

	attr.type = PERF_TYPE_HARDWARE;
	attr.config = PERF_COUNT_HW_CPU_CYCLES;
	perfdev = syscall(__NR_perf_event_open, &attr, 0, -1, -1, 0);

	if (perfdev <= 0) {
		printf("perf_event_open() with PERF_COUNT_HW_CPU_CYCLES failed: %s\n", strerror(errno));
		return;
	}

	test_cpucycles_function(cpucycles_perf, "PERF_COUNT_HW_CPU_CYCLES");

	return;
}

void test_PERF_COUNT_HW_REF_CPU_CYCLES() {
	static struct perf_event_attr attr;
	int perfdev;

	attr.type = PERF_TYPE_HARDWARE;
	attr.config = PERF_COUNT_HW_REF_CPU_CYCLES;
	perfdev = syscall(__NR_perf_event_open, &attr, 0, -1, -1, 0);

	if (perfdev <= 0) {
		printf("perf_event_open() with PERF_COUNT_HW_REF_CPU_CYCLES failed: %s\n", strerror(errno));
		return;
	}

	test_cpucycles_function(cpucycles_perf, "PERF_COUNT_HW_REF_CPU_CYCLES");

	return;
}


void test_cpucycles() {
	test_cpucycles_function(cpucycles, "RDTSC");
}


int main(int argc, char *argv[]) {

	test_PERF_COUNT_HW_CPU_CYCLES();
	test_PERF_COUNT_HW_REF_CPU_CYCLES();
#ifndef __arm__
	test_cpucycles();
#endif
}

