//______________________________________________________________________________
// Taken from the ethos software distribution.
// Stripped from managment and network code
// Replaced nacl with sodium
//
// Janosch Rux -- Feb 2014
// Jon Solworth -- Aug 2014
//______________________________________________________________________________

#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <memory.h>
#include <sodium.h>

#include "minimalt_puzzle.h"

uint64_t mixinGetTimeOfDay() {
	static uint64_t i = 100000000000;
	return i++;
}


void memzero(void *p, size_t count) {
	memset(p, 0, count);
}

uchar *
netWrite64(uchar *buff, uint64 v)
{
    memcpy(buff, (void *)&v, sizeof(v));
    buff += sizeof(v);

    return buff;
}

uchar *
netRead64(uchar *buff, uint64 *v) 
{
    memcpy(v, buff, sizeof(uint64));
    buff += sizeof(uint64);

    return buff;
}

static inline int cryptoHash(uchar *hash, const uchar *plainText, unsigned long long plainTextLen)
{
    return crypto_hash_sha512(hash, plainText, plainTextLen);
}

static inline int cryptoSecretBox(uchar *c,
                const uchar *m,
                unsigned long long mlen,
                const uchar *n,
                const uchar *k)
{
    return crypto_secretbox(c, m, mlen, n, k);
}


static inline int cryptoSecretBoxOpen(uchar *m,
                    const uchar *c,
                    unsigned long long clen,
                    const uchar *n,
                    const uchar *k)
{
    return crypto_secretbox_open(m, c, clen, n, k);
}


// one puzzle encryptionKey per server
static CryptoKey _encryptionKey[SymmetricKeySize];

uint64
_byteswap64(uint64 input)
{
#ifdef __GNUC__
    return __builtin_bswap64(input);
#else
    input = (input & 0x00000000FFFFFFFF) << 32 | (input & 0xFFFFFFFF00000000) >> 32;
    input = (input & 0x0000FFFF0000FFFF) << 16 | (input & 0xFFFF0000FFFF0000) >> 16;
    input = (input & 0x00FF00FF00FF00FF) << 8  | (input & 0xFF00FF00FF00FF00) >> 8;
    return input;
#endif
}

//______________________________________________________________________________
// build the server-encrypted token, consisting of:
//    1. server key
//    2. client key
// encrypted with a secret, k, known only to the server
//______________________________________________________________________________
static
void
_buildR(CryptoKey *clientKey, CryptoKey *serverKey,  uint64 n, uchar *k, uchar *r)
{
    uchar rPadded[PuzzleEncryptedPaddedSize];
    uchar m[PuzzleEncryptedPaddedSize];
    uchar nc[crypto_secretbox_NONCEBYTES];

    // prepare m
    uchar *ptr = m;
    memzero(ptr, crypto_secretbox_ZEROBYTES);
    ptr +=  crypto_secretbox_ZEROBYTES;
    memcpy(ptr, clientKey, EncryptionPublicKeySize);
    ptr +=  EncryptionPublicKeySize;
    memcpy(ptr, serverKey, EncryptionPublicKeySize);

    // prepare nc
    memzero(nc, crypto_secretbox_NONCEBYTES);
    memcpy(nc, &n, sizeof(uint64));

    // compute r 
    cryptoSecretBox(rPadded, m, PuzzleEncryptedPaddedSize, nc, k);

    // copy & get rid of box zeros
    memcpy(r, rPadded+crypto_secretbox_BOXZEROBYTES, PuzzleEncryptedSize);
}


//______________________________________________________________________________
//  Zero out the low order -b- bits
//______________________________________________________________________________
static
void
_zeroBits(uint32 b, uchar *r)
{

    uint64 value;
    uchar *ptr = r + PuzzleEncryptedSize - sizeof(uint64);
    netRead64(ptr, &value);

    value &= ~0ULL << b;

    netWrite64(ptr, value);
}

void puzzleInit() {
    // create a symmetric key with which to create puzzle 
    memset(_encryptionKey, 42, SymmetricKeySize);
}


//______________________________________________________________________________
// geenerate a new puzzle
//______________________________________________________________________________
bool
puzzleGenerate(CryptoKey *clientKey,
	       CryptoKey *serverKey,
	       uint32     b,
	       Puzzle     *puzzle)
{

    puzzle->nonce = mixinGetTimeOfDay();
    puzzle->missingBits = b;
    _buildR(clientKey, serverKey, puzzle->nonce, _encryptionKey, puzzle->encryptedData);
    cryptoHash(puzzle->hash, puzzle->encryptedData, PuzzleEncryptedSize);
    _zeroBits(b, puzzle->encryptedData);
    return true;
}


static
bool
_verifyNonce(uint64 n)
{
    // Has to be solved within 5 seconds
    return (mixinGetTimeOfDay() - 5000000000) < n;
}


static
bool
_verifyBoxContent(uchar *m, CryptoKey *clientKey, CryptoKey *serverKey)
{
    bool ok  = !memcmp(m, clientKey, EncryptionPublicKeySize);
    ok      &= !memcmp(m + EncryptionPublicKeySize, serverKey, EncryptionPublicKeySize);

    return ok;
} 

//______________________________________________________________________________
// Check that the puzzle has indeed been solved.  Used only at the server-side,
// since k is needed.
//______________________________________________________________________________
bool
puzzleVerify(PuzzleSolution *puzzle, CryptoKey *clientKey, CryptoKey *serverKey)
{

    bool status;

    uchar c[PuzzleEncryptedPaddedSize];    // ciphertext
    uchar nc[crypto_secretbox_NONCEBYTES]; // nonce
    uchar m[PuzzleEncryptedPaddedSize];    // unencrypted

    if (!_verifyNonce(puzzle->nonce))
	{
	    return false;
	}
 
    memzero(nc, crypto_secretbox_NONCEBYTES);
    memcpy(nc, &puzzle->nonce, sizeof(puzzle->nonce));

    memzero(c, crypto_secretbox_BOXZEROBYTES);
    memcpy(c+crypto_secretbox_BOXZEROBYTES, puzzle->encryptedData, PuzzleEncryptedSize);

    memzero(m, PuzzleEncryptedPaddedSize);

    if (cryptoSecretBoxOpen(m, c, PuzzleEncryptedPaddedSize, nc, _encryptionKey) == -1) 
	{
		return false;
	}

    return _verifyBoxContent(m+crypto_secretbox_ZEROBYTES, clientKey, serverKey);
}


//______________________________________________________________________________
//  brute force solution to the puzzle
//______________________________________________________________________________
bool
puzzleSolve(Puzzle *challenge, uint maxPuzzleBits)
{

    uchar *targetHash = challenge->hash;
    uchar          *q = challenge->encryptedData;
    uint32       bits = challenge->missingBits;
    uchar       qHash[HashSize];

    // limit bits to less than 64,
    // ensure puzzle not too large computationally
    if (bits > maxPuzzleBits)
	{
	    return false;
	}

    uchar *ptr = q + PuzzleEncryptedSize - sizeof(uint64);

    // get any counter bits in q
    uint64 mask   = (1ULL << bits) - 1;
    uint64 counter;
    netRead64(ptr, &counter);
    counter &= ~mask; // force low order bits to zero
    uint64 bound = counter + mask;
    for ( ; counter <= bound; counter++) 
	{
	    netWrite64(ptr, counter);
	    cryptoHash(qHash, q, PuzzleEncryptedSize);
	    if (memcmp(qHash, targetHash, HashSize) == 0)
		{
		    return true;
		}
	
	}
    return false;
}
